/*    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2013-
 *    Salvatore Ruggieri
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package rst.table;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Random;

import rst.table.Column;
import rst.table.Table;
import rst.util.data.Bitmap;
import rst.util.data.Pair;
import rst.util.data.PairIntInt;

/** A subset of a rst.table.Table. */
public class TableSubset {
	/** Row positions in the subset. */
	public ArrayList<Integer> rows;
	
	/** Constructor. */
	public TableSubset() {
		rows = new ArrayList<Integer>();
	}
	
	/** Constructor. */
	public TableSubset(Table tb) {
		this();
		int size = tb.size();
		rows.ensureCapacity(size);
		for(int i=0; i<size; ++i)
			rows.add(i);
	}
	
	/** Constructor. */
	public TableSubset(ArrayList<Integer> rows) {
		this.rows = rows;
	}
	
	/** Random split. */
	public Pair<TableSubset, TableSubset> splitRandom(double perc) {
		int tot = rows.size();
		int to_sel = (int) Math.ceil(perc*tot);
		Random r = new Random();
		TableSubset left = new TableSubset();
		TableSubset right = new TableSubset();
		int i = 0;
		while(tot>0) {
			int ran = r.nextInt(tot);
			if(ran<to_sel) {
				left.rows.add(rows.get(i));
				to_sel--;
			} else
				right.rows.add(rows.get(i));
			++i;	
			--tot;
		}
		return new Pair<TableSubset, TableSubset>(left, right);
	}
	
	/** Size of subset. */
	public int size() {
		return rows.size();
	}
	
	/** Conversion to bitmap of nBits. */
	public Bitmap toBitmap(int nBits) {
		Bitmap b = new Bitmap(nBits);
		for(int i: rows)
			b.set(i);
		return b;
	}
	
	/** Get smallest-highest index appearing in the subset for each column. */
	public PairIntInt[] getRangeIdx(Table tb) {
		int ncols = tb.columns.size();
		PairIntInt range[] = new PairIntInt[ncols];
		for(int i=0; i<ncols; ++i) {
			Column col = tb.columns.get(i);
			int minIdx = Integer.MAX_VALUE;
			int maxIdx = Integer.MIN_VALUE;
			for(int e : rows) {
				int idx = col.index[e];
				if(idx < minIdx)
					minIdx = idx;
				if(idx > maxIdx)
					maxIdx = idx;
			}
			range[i] = new PairIntInt(minIdx, maxIdx);
		}
		return range;
	}
	
	/** Split subset on a given attribute and mid index. */
	public TableSubset split(Table tb, int att, int mid, boolean left) {
		TableSubset res = new TableSubset();
		Column col = tb.columns.get(att);
		if( left ) {
			for(int e : rows) { 
				int val = col.index[e];
				if( val <= mid )
					res.rows.add(e);
			}
		} else {
			for(int e : rows) {
				int val = col.index[e];
				if( val > mid )
					res.rows.add(e);
			}
		}
		return res;
	}

	/** Return distribution of an attribute' indexes in the subset. */
	public int[] getDistribution(Table tb, int att) {
		Column col = tb.columns.get(att);
		int nval = col.getLastIndex() + 1;
		// SELECT val, COUNT(*) FROM subset GROUP BY val
		int []group = new int[nval];
		Arrays.fill(group, 0);
		for(int v : rows)
			group[col.index[v]]++;
/*		System.out.print("Dist: ");
		for(int i=0; i<nval; ++i)
			System.out.print(group[i]+" ");
		System.out.println(); */
		return group;
	}

	/** Return median value of an attribute in the subset, given its distribution. */
	public int getMedian(int[] group) {
		int n = rows.size();
		int half = (n+1)/2;
		int median = 0;
		int sum  = group[median];
		while( sum < half )
			sum += group[++median];
		if( sum < n )
			return median;
		if(group[median]==sum)
			return -(median+1);
		while( group[--median] == 0) ;
		return median;
	}
	
	/** Return normalized range. */
	public double getNormalizedRange(int[] group) {
		int min = 0;
		for(int i=0; i<group.length; ++i) {
			if(group[i]==0)
				continue;
			min = i;
			break;
		}
		int max = -1;
		for(int i=group.length-1; i>=0; --i) {
			if(group[i]==0)
				continue;
			max = i;
			break;
		}
		return (max-min+1)/(double)group.length;
	}

	/** Return median value of an attribute in the subset. */
	public int getMedian(Table tb, int att) {
		return getMedian(getDistribution(tb, att) );
	}
}
	
