/*    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2013-
 *    Salvatore Ruggieri
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package rst.util.data;

import java.io.Serializable;

/** A bitmap data structure. */
@SuppressWarnings("serial")
public class Bitmap implements Comparable<Bitmap>, Serializable
{
	/** Bits stored in integers. */
	protected int[] val;
	/** Number of bits. */
	protected int nBits;
	
	/** Default size of a bitmap */
	/** Constructor. */
	public Bitmap(int nBits) {
		int nint = 1+ ((nBits-1) >> 5);
		val = new int[nint];
		this.nBits = nBits;
	}
	/** Constructor. Re-use storage. */
	private Bitmap(int[] val, int nBits) {
		this.val = val;		
		this.nBits = nBits;
	}
	/** Copy constructor. */
	public Bitmap(Bitmap b) {
		int nint = b.val.length;
		val = new int[nint];
		int[] bval = b.val;
		for(int i=0; i<nint; ++i)
			val[i] = bval[i];
		this.nBits = b.nBits;
	}
	/** Clone. */
	public Bitmap clone() {
		return new Bitmap(this);
	}
	/** Size of the bitmap. */
	public int size() {
		return nBits;
	}
	/** Set to 1 a bit position. */
	public void set(int pos)
	{
		int bytep = pos >> 5;
		int relp = pos & 31;
		val[bytep] |= (1 << relp);
	}
	/** Set to 0 a bit position. */
	public void unset(int pos)
	{
		int bytep = pos >> 5;
		int relp = pos & 31;
		val[bytep] &= ~(1 << relp );
	}
	/** Set to 0 all bits. */
	public void unsetall()
	{
		int nint = val.length;
		for(int i=0; i<nint; ++i)
			val[i] = 0;
	}
	/** Set to 1 all bits. */
	public void setall()
	{
		int nint = val.length;
		for(int i=0; i<nint; ++i)
			val[i] = ~0;
		// reset bits after nBits - can be made more efficient
		int totBits = nint << 5;
		for(int i=nBits; i<totBits; ++i)
			unset(i);
	}
	/** Set to 1 all and only all bits passed in an array of positions. */
	public void setall(int[] items, int size) {
		for(int i=0; i<size; ++i)
			set(items[i]);
	}
	/** Get the value of a bit. */
	public boolean get(int pos)
	{
		int bytep = pos >> 5;
		int relp = pos & 31;
		return (val[bytep] & (1 << relp )) != 0;
	}
	/** Count number of bits set to 1. */
	public int bitCount()
	{
		int nint = val.length;
		int sum = bitCount(val[0]);
		for(int i=1; i<nint; ++i)
			sum += bitCount(val[i]);
		return sum;
	}
	/** Count number of bits set to 1 in the passed parameter. */
	public static int bitCount(int n)
	{
		int count = 0;
	    while (n != 0) {
	      n &= (n - 1);
	      count++;
	    }
	    return count;
	}
	/** Get pos-th bit set to 1. */
	public int getByRank(int pos) {
		int bytep = 0;
		int found = 0;
		int count = -1;
		while( count != pos ) {
			found = bytep * 32;
			int n = val[bytep];
			bytep++;
			for(int i=0, k=1; i<32; ++i) {
		      if ( (n & k) != 0)
		    	  if( ++count == pos )
		    		  break;
		      k <<= 1;
		      ++found;
		    }
		}
	    return found;
	}
	/** Or of bitmaps. */	
	public Bitmap or(Bitmap b) {
		int nint = val.length;
		int[] val2 = new int[nint];		
		int[] bval = b.val;
		for(int i=0; i<nint; ++i)
			val2[i] = val[i] | bval[i];
		return new Bitmap(val2, nBits);
	}
	/** Not of this bitmaps. */	
	public Bitmap not() {
		int nint = val.length;
		int[] val2 = new int[nint];
		for(int i=0; i<nint; ++i)
			val2[i] = ~val[i];
		Bitmap result = new Bitmap(val2, nBits);
		// reset bits after nBits
		int totBits = nint << 5;
		for(int i=nBits; i<totBits; ++i)
			result.unset(i);
		return result;
	}
	/** And of bitmaps. */	
	public Bitmap and(Bitmap b) {
		int nint = val.length;
		int[] val2 = new int[nint];		
		int[] bval = b.val;
		for(int i=0; i<nint; ++i)
			val2[i] = val[i] & bval[i];
		return new Bitmap(val2, nBits);
	}
	/** And of bitmaps replacing current. */	
	public void and_inplace(Bitmap b) {
		int nint = val.length;
		int[] bval = b.val;
		for(int i=0; i<nint; ++i)
			val[i] = val[i] & bval[i];
	}
	/** Size of and of bitmap. */
	public int and_size(Bitmap b) {
		int nint = val.length;
		int[] bval = b.val;
		int sz = 0;
		for(int i=0; i<nint; ++i)
			sz += bitCount(val[i] & bval[i]);
		return sz;
	}
	/** Difference of bitmaps. */	
	public Bitmap minus(Bitmap b) {
		int nint = val.length;
		int[] val2 = new int[nint];		
		int[] bval = b.val;
		for(int i=0; i<nint; ++i)
			val2[i] = val[i] & ~bval[i];
		return new Bitmap(val2, nBits);
	}
	/** Equality test. */
	public boolean equals(Bitmap o) {
		int nint = val.length;
		for(int i=0; i<nint; ++i)
			if( val[i] != o.val[i] )
				return false;
		return true;
	}
	/** Compare bitmaps. */
	public int compareTo(Bitmap t) {
		int nint = val.length;
		int[] tval = t.val;
		for(int i=0; i<nint; ++i) {
			int v = val[i];
			int tv = tval[i];
			if( v < tv )
				return -1;
			if( v > tv )
				return 1;
		}
		return 0; 
	}
	/** Array of bits set. */
	public int[] toArray() {
		int len = this.bitCount();
		int []result = new int[len];
		for(int i=0; i<len-1; ++i)
			result[i] = this.getByRank(i); // implementation can be improved
		return result;		
	}
	/** String representation. */
	public String toString()
	{
		String res = "";
		int len = this.bitCount();
		for(int i=0; i<len-1; ++i)  // implementation can be improved
			res = res + this.getByRank(i) + " ";
		if( len > 0 )
			res = res + this.getByRank( len - 1 );
		return res;
	}
	/** Test method. */
	public static void main(String[] args) {
		final int sz = 96;
		Bitmap b = new Bitmap(sz);
		Bitmap c = new Bitmap(sz);
		b.set(19);
		b.set(28);
		
		c.set(28);
		for(int i=0;i<sz;++i)
			System.out.println( "" + i + " " + b.get(i) );
		int bc = b.bitCount();
		System.out.println( "Bitcount b = " + bc );
		for(int i=0;i<bc;++i)
			System.out.println( "" + i + " " + b.getByRank(i) );
		bc = c.bitCount();
		System.out.println( "Bitcount c = " + bc );
		for(int i=0;i<bc;++i)
			System.out.println( "" + i + " " + c.getByRank(i) );
		System.out.println("b.compareTo(c) = " + b.compareTo(c));
	}

}
